-----------------------------------------------------------------------------
--
-- Tag				:= { Statement } @
-- Statement		:= literal | "[" RValue "]"
-- RValue			:= Relation
-- Relation			:= Expression [ ("==" | "~=" | ">" | "<" | "<=" | ">=") Expression ]
-- Expression		:= Term [ ("+" | "-") Term ]
-- Factor			:= Chain [ ("*" | "/") Chain ]
-- Chain			:= Primary { ":" Primary }
-- Primary			:= ConstantNumber | ConstantString | Identifier ["(" ParameterList ")"] | ( RValue )
-- ParameterList	:= RValue {, RValue }
--
-----------------------------------------------------------------------------

AloftTag = { }
AloftTagMethods = { }
AloftTagDirectAccessData = { isBoss = true, level = true, name = true, health = true, maxHealth = true, healthBarValue = true, healthBarMaxValue = true }
AloftTagData = 
{ 
	health 			= { events = "Aloft:OnHealthBarValueChanged"},
	healthBarValue 	= { events = "Aloft:OnHealthBarValueChanged"},
	maxHealth 		= { events = "Aloft:OnHealthBarValueChanged"},
}

-----------------------------------------------------------------------------

AloftTagMethods.angle = function(aloftData, value) return value and value ~= "" and string.format("<%s>", value) or nil end
AloftTagMethods.bracket = function(aloftData, value) return value and value ~= "" and string.format("(%s)", value) or nil end
AloftTagMethods.brace = function(aloftData, value) 	return value and value ~= "" and string.format("{%s}", value) or nil end

AloftTagMethods.color = function(aloftData, value, r, g, b)
	return value and string.format("|cff%02x%02x%02x%s|r", r*255, g*255, b*255, value or "") or nil
end

AloftTagMethods.colorIf = function(aloftData, value, condition, r, g, b)
	return value and condition and string.format("|cff%02x%02x%02x%s|r", r*255, g*255, b*255, value or "") or value
end

AloftTagMethods.healthBarDeficit = function(aloftData, value)
	return value and (aloftData.healthBarMaxValue - aloftData.healthBarValue) or nil
end
AloftTagData.healthBarDeficit = { events = "Aloft:OnHealthBarValueChanged" }

function AloftTagMethods.healthDeficit(aloftData, value)
	if value and aloftData.health and aloftData.maxHealth then
		return aloftData.maxHealth - aloftData.health
	else
		return nil
	end
end
AloftTagData.healthDeficit = { events = "Aloft:OnHealthBarValueChanged" }

AloftTagMethods.healthPercent = function(aloftData, value)
	return value and floor(100*aloftData.healthBarValue / aloftData.healthBarMaxValue) or nil
end
AloftTagData.healthPercent = { events = "Aloft:OnHealthBarValueChanged" }

function AloftTagMethods.hideKnownHealth(aloftData, value)
	return not aloftData.health and value or nil
end

AloftTagMethods.isGroup = function(aloftData, value)
	return aloftData.healthBarMaxValue ~= 100 and value or nil
end

AloftTagMethods.isNotGroup = function(aloftData, value)
	return aloftData.healthBarMaxValue == 100 and value or nil
end

AloftTagMethods.hexColor = function(aloftData, value, colorValue)
	return value and string.format("|cff%s%s|r", colorValue, value or "") or nil
end

AloftTagMethods.hexColorIf = function(aloftData, value, condition, colorValue)
	return value and condition and string.format("|cff%s%s|r", colorValue, value or "") or value
end

AloftTagMethods.hideIf = function(aloftData, value, condition)
	return not condition and value or nil
end

AloftTagMethods.hideZero = function(aloftData, value)
	return value ~= 0 and value or nil
end

AloftTagMethods.isFriendly = function(aloftData, value)
	return (aloftData.type == "friendlyPlayer" or aloftData.type == "friendlyNPC") and value
end

AloftTagMethods.isFriendlyNPC = function(aloftData, value)
	return aloftData.type == "friendlyPlayer" and value
end

AloftTagMethods.isFriendlyPlayer = function(aloftData, value)
	return aloftData.type == "friendlyPlayer" and value
end

AloftTagMethods.isHostile = function(aloftData, value)
	return aloftData.type == "hostile" and value
end

AloftTagMethods.isNeutral = function(aloftData, value)
	return aloftData.type == "neutral" and value
end

AloftTagMethods.isMaxHealth = function(aloftData, value)
	return aloftData.healthBarValue == aloftData.healthBarMaxValue and value or nil
end
AloftTagData.isMaxHealth = { events = "Aloft:OnHealthBarValueChanged" }

AloftTagMethods.lower = function(aloftData, value)
	if type(value) == "string" then
		return value:lower()
	end
	return value
end

AloftTagMethods.negate = function(aloftData, value)
	if type(value) == "number" then
		return -value
	end
	return value
end

AloftTagMethods["not"] = function(aloftData, value, condition)
	return not condition and value
end

AloftTagMethods.percent = function(aloftData, value)
	if type(value) == "number" then
		return value.."%"
	end
	return value
end

AloftTagMethods.prefix = function(aloftData, value, prefix)
	return value and prefix..value or nil
end

AloftTagMethods.rgb = function(aloftData, value, r, g, b)
	return value and string.format("|cff%02x%02x%02x%s|r", r, g, b, value or "") or nil
end

AloftTagMethods.rgbIf = function(aloftData, value, condition, r, g, b)
	return value and condition and string.format("|cff%02x%02x%02x%s|r", r, g, b, value or "") or value
end

AloftTagMethods.short = function(aloftData, value)
	if type(value) == "number" then
		if value < 10000 then
			-- Force it to an integer
			return string.format("%d", value)
		elseif value < 100000 then
			return string.format("%.1fk", value / 1000)
		elseif value < 1000000 then
			return string.format("%dk", value / 1000)
		else
			return string.format("%.2fm", value / 1000000)
		end
	end
	return value
end

AloftTagMethods.suffix = function(aloftData, value, suffix)
	return value and value..suffix or nil
end

AloftTagMethods.upper = function(aloftData, value)
	if type(value) == "string" then
		return value:upper()
	end
	return value
end

AloftTagMethods.white 	= function(aloftData, value) return value and ("|r%s"):format(value) or nil end
AloftTagMethods.red 	= [[hexColor("ff0000")]]
AloftTagMethods.green 	= [[hexColor("00ff00")]]
AloftTagMethods.blue 	= [[hexColor("0000ff")]]
AloftTagMethods.cyan 	= [[hexColor("00ffff")]]
AloftTagMethods.fuchsia = [[hexColor("ff00ff")]]
AloftTagMethods.yellow 	= [[hexColor("ffff00")]]
AloftTagMethods.gray 	= [[hexColor("afafaf")]]
AloftTagMethods.difficultyColor = function(aloftData, value) return value and ("|cff%02x%02x%02x%s|r"):format(aloftData.levelTextR*255, aloftData.levelTextG*255, aloftData.levelTextB*255, value) or nil end

-----------------------------------------------------------------------------

local parseTag
local parseTagOffset
local tagMethods = { }
local numeric = false

local function GetByte()
	while parseTag:byte(parseTagOffset) == 32 do 
		parseTagOffset = parseTagOffset + 1
	end 
	return parseTag:byte(parseTagOffset) or 0
end

local function AdvanceByte()
	parseTagOffset = parseTagOffset + 1
end

-----------------------------------------------------------------------------

local function TagErrorStringMethod() return "!!Tag Error!!" end
local function TagError(x) Aloft:Print(x) end

local function EmptyTable(table) for k in pairs(table) do table[k] = nil end end

function AloftTag:CoerceToNumber(text)
	if not text or text == "" then return "0" end
	if numeric then return text end
	return string.format("(%s or 0)", text)
end

function AloftTag:ParseParameterList(dataTable)
	if GetByte() == 40 then -- byte == '('
		AdvanceByte()
		local byte = GetByte()
		if byte == 41 then 
			AdvanceByte()
			return ""
		end

		local parameterListString = string.format(", %s", self:ParseRValue(dataTable))
		
		while parseTagOffset <= #parseTag and GetByte() == 44 do -- byte == ','
			AdvanceByte()
			parameterListString = string.format("%s, %s", parameterListString, self:ParseRValue(dataTable))
		end

		if GetByte() == 41 then -- byte == ')'
			AdvanceByte()
		else
			dataTable.error = true
			TagError("Missing end of parameter list \")\"")
		end
		
		return parameterListString
	else
		return ""
	end
end

function AloftTag:ParsePrimary(dataTable, functionString)
	-- Either an integer, a string or an LValue (value or function call)
	numeric = false
	
	local byte = GetByte()
	if byte >= 48 and byte <= 59 then -- byte >= '0' && byte <= '9'
		local startOffset = parseTagOffset
		local endOffset = startOffset
		local decimalPointFound = false
		while true do
			endOffset = endOffset + 1
			byte = parseTag:byte(endOffset) or 0
			if byte == 46 then -- '.'
				if decimalPointFound then
					break
				else
					decimalPointFound = true
				end
			elseif byte < 48 or byte > 59 then
				break
			end
		end
		parseTagOffset = endOffset
		if functionString then
			return string.format("(%s and %s or nil)", functionString, parseTag:sub(startOffset, endOffset-1))
		else
			numeric = true
			return parseTag:sub(startOffset, endOffset-1)
		end
	elseif byte == 34 then	-- byte == '\"'
		local startOffset = parseTagOffset+1
		local endOffset = startOffset + 1
		while true do
			if endOffset > #parseTag then
				TagError("Missing closing quote")
				dataTable.error = true
				return ""
			end
			byte = parseTag:byte(endOffset) or 0
			if byte == 34 then -- byte == '\"'
				break
			elseif byte == 92 then	-- byte == '\\'
				endOffset = endOffset + 2
			else
				endOffset = endOffset + 1
			end
		end
		parseTagOffset = endOffset+1
		if functionString then
			return string.format("%s and %q", functionString, parseTag:sub(startOffset, endOffset-1))
		else
			return string.format("%q", parseTag:sub(startOffset, endOffset-1))
		end
		
	elseif (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) then	-- byte >= 'A' && byte <= 'Z' || byte >= 'a' && byte <= 'z'
		local startOffset = parseTagOffset
		local endOffset = startOffset
		while (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) do
			endOffset = endOffset + 1
			byte = parseTag:byte(endOffset) or 0
		end
		parseTagOffset = endOffset
		local identifier = parseTag:sub(startOffset, endOffset-1)
		local data = AloftTagData[identifier]
		if data then
            if data.events then
				if type(data.events) == "string" then
                    dataTable.events[data.events] = true
				elseif type(data.events) == "table" then
					for _,v in pairs(data.events) do
						dataTable.events[v] = true
					end
                end
            end
            if data.data then
                if type(data.data) == "string" then
                    dataTable.data[data.data] = true
				elseif type(data.data) == "table" then
					for _,v in pairs(data.data) do
						dataTable.data[v] = true
					end
                end
            end
		end
		if AloftTagDirectAccessData[identifier] then
			if functionString then
				return functionString.." and aloftData."..identifier
			else
				return "aloftData."..identifier
			end
		elseif AloftTagMethods[identifier] then
			if type(AloftTagMethods[identifier]) == "string" then
				local tag, offset = parseTag, parseTagOffset
				parseTag, parseTagOffset = AloftTagMethods[identifier], 1
				local returnValue = self:ParsePrimary(dataTable, functionString)
				parseTag, parseTagOffset = tag, offset
				return returnValue
			else
				tagMethods[identifier] = true	-- this marks the method as required, and it's inserted as an upvalue
				local parameterList = self:ParseParameterList(dataTable)
				return string.format("tagMethod%s(aloftData, %s%s)", identifier, functionString or "true", parameterList)
			end
		else
			TagError("Invalid tag: "..identifier)
			dataTable.error = true
			return ""
		end
	elseif (byte == 40) then -- '('
		AdvanceByte()
		local result = self:ParseRValue(dataTable)
		local byte = GetByte()
		if byte == 41 then
			AdvanceByte()
			return string.format("(%s)", result)
		else
			TagError("Missing \")\"")
			dataTable.error = true
			return ""
		end
	end
end

function AloftTag:ParseChain(dataTable)
	local functionString = self:ParsePrimary(dataTable)
	while GetByte() == 58 do	-- while tag[offset] == ':'
		AdvanceByte()
		functionString = self:ParsePrimary(dataTable, functionString)
	end
	return functionString
end

function AloftTag:ParseFactor(dataTable)
	local result = self:ParseChain(dataTable)
	local byte = GetByte()
	while byte == 42 or byte == 47 do	-- '*' or '/'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		result = string.format("%s %s %s", result, string.char(byte), self:CoerceToNumber(self:ParseFactor(dataTable)))
		numeric = true
		byte = GetByte()
	end
	return result
end

function AloftTag:ParseExpression(dataTable)
	local result = self:ParseFactor(dataTable)
	local byte = GetByte()
	while byte == 43 or byte == 45 do	-- '+' or '-'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		result = string.format("%s %s %s", result, string.char(byte), self:CoerceToNumber(self:ParseFactor(dataTable)))
		numeric = true
		byte = GetByte()
	end
	return result
end

function AloftTag:ParseRelation(dataTable)
	local result = self:ParseExpression(dataTable)

	local byte = GetByte()
	local byte2 = parseTag:byte(parseTagOffset+1)
	if byte == 61 then	-- '='
		AdvanceByte()
		if byte2 == 61 then
			AdvanceByte()
			return result.."=="..self:ParseExpression(dataTable)
		else
			TagError("Unexpected \"=\"")
			dataTable.error = true
			return ""
		end
	elseif byte == 126 then -- '~'
		AdvanceByte()
		if byte2 == 61 then
			AdvanceByte()
			return result.."~="..self:ParseExpression(dataTable)
		else
			TagError("Unexpected \"=\"")
			dataTable.error = true
			return ""
		end
	elseif byte == 60 then -- '<'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		if byte2 == 61 then
			AdvanceByte()
			return string.format("%s <= %s", result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		else
			return string.format("%s < %s", result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		end
	elseif byte == 62 then -- '>'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		if byte2 == 61 then
			AdvanceByte()
			return string.format("%s >= %s", result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		else
			return string.format("%s > %s", result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		end
	else
		return result
	end
end

function AloftTag:ParseRValue(dataTable)
	return self:ParseRelation(dataTable)
end

-- Initialise the compile cache with an empty tag, something which would otherwise need some extra special-cased code
local compileCache =
{
	[""] = { data = { }, events = { }, tag = "", method = function() return "" end },
}

local headerTable = { }
local functionStringList = { }
function AloftTag:CreateTag(tag)
	parseTag = tag or ""

	if compileCache[parseTag] then
		return compileCache[parseTag]
	end

	parseTagOffset = 1
	local dataTable = { data = { }, events = { }, tag = parseTag }

	table.insert(functionStringList, "") -- it's replaced with all the locals required later
	table.insert(functionStringList, "return function(aloftData) return ")

	while parseTagOffset <= #parseTag and not dataTable.error do
		if parseTagOffset > 1 then
			table.insert(functionStringList, "..")
		end
		if parseTag:byte(parseTagOffset) == 91 then	-- if tag[offset] == '['
			AdvanceByte()
			table.insert(functionStringList, string.format([[(%s or "")]], self:ParseRValue(dataTable) or ""))
			if GetByte() == 93 then -- if tag[offset] == ']'
				AdvanceByte()
			elseif not dataTable.error then
				TagError("Tag not closed - missing \"]\"")
				dataTable.error = true
			end
		else
			-- literal section
			local startOffset = parseTagOffset
			local endOffset = startOffset+1
			while endOffset <= #parseTag and parseTag:byte(endOffset) ~= 91 do
				endOffset = endOffset + 1
			end
			table.insert(functionStringList, string.format("%q", parseTag:sub(startOffset, endOffset-1)))
			parseTagOffset = endOffset
		end
	end

	if dataTable.error then
		dataTable.method = TagErrorStringMethod
	else
		for k in pairs(tagMethods) do
			table.insert(headerTable, string.format("local tagMethod%s = AloftTagMethods[\"%s\"]\n", k, k))
		end
		functionStringList[1] = table.concat(headerTable)
		EmptyTable(headerTable)
		table.insert(functionStringList, " end")
		local functionString = table.concat(functionStringList)
--[[
		Aloft:Debug(string.format("Tag: %s", parseTag))
		for event in pairs(dataTable.events) do
			Aloft:Debug(string.format(" - event: %s", event))
		end
		for data in pairs(dataTable.data) do
			Aloft:Debug(string.format(" - data: %s", data))
		end
		Aloft:Debug("--- Compiled method ---\n"..functionString)
--]]--
		dataTable.method = loadstring(functionString)()
		compileCache[parseTag] = dataTable
	end

	EmptyTable(tagMethods)
	EmptyTable(functionStringList)

	return dataTable
end

-----------------------------------------------------------------------------
