-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AloftFrame")

-----------------------------------------------------------------------------

L:RegisterTranslations("enUS", function() return {
	["Enable Mouse"] = true,
	["Allows left click to target and right click to interact"] = true,

	["Frame"] = true,
	["Frame options"] = true,
	["Width"] = true,
	["Sets the width of the frame"] = true,
	["Height"] = true,
	["Sets the height of the frame"] = true,

	["Packing Width"] = true,
	["Sets the minimum horizontal distance between frames"] = true,
	["Packing Height"] = true,
	["Sets the minimum vertical distance between frames"] = true,

	["Background Alpha"] = true,
	["Sets the alpha for the background of the nameplates"] = true,
	["Background Colors"] = true,
	["Customize colors"] = true,
	
	["Color by Class"] = true,
	["Sets the background color according to class"] = true,
	["Color Hostile by Class"] = true,
	["Sets the background color of hostile players according to class"] = true,
	["Set All"] = true,
	["Sets the background color for all types"] = true,

	["Set All Friendly"] = true,
	["Sets the background color for all friendly types"] = true,
	["Set All Hostile"] = true,
	["Sets the background color for all hostile types"] = true,

	["Pet"] = true,
	["Sets the background color for your pet"] = true,
	["Friendly Players"] = true,
	["Sets the background color for friendly players"] = true,
	["Friendly Pets"] = true,
	["Sets the background color for friendly pets"] = true,
	["Friendly NPCs"] = true,
	["Sets the background color for friendly NPC Bosses"] = true,
	["Friendly Boss"] = true,
	["Sets the background color for friendly NPCs"] = true,
	["Hostile Players"] = true,
	["Sets the background color for hostile players"] = true,
	["Hostile Pets"] = true,
	["Sets the background color for hostile pets"] = true,
	["Hostile NPCs"] = true,
	["Sets the background color for hostile units"] = true,
	["Hostile Boss"] = true,
	["Sets the background color for hostile boss units"] = true,
	["Neutral Units"] = true,
	["Sets the background color for neutral units"] = true,
	["Group Pets"] = true,
	["Sets the background color for group pets"] = true,

	["Advanced"] = true,
	["Options for expert users"] = true,
	["Color Format"] = true,
	["Color tag"] = true,
	["<Any tag string>"] = true,
	
	["Border Color"] = true,
	["Sets the border color of the nameplate"] = true,
	["Border Style"] = true,
	["Sets the style of the nameplate border"] = true,
	
	["NONE"] = "None", 
	["STANDARD"] = "Standard", 
	["SQUARE"] = "Square",
} end)

-----------------------------------------------------------------------------

local borderData =
{
	["NONE"] 		= { texture = "", size = 0, inset = 0 },
	["STANDARD"] 	= { texture = "Interface\\Tooltips\\UI-Tooltip-Border", size = 16, inset = 5 },
	["SQUARE"]	 	= { texture = "Interface\\DialogFrame\\UI-DialogBox-Border", size = 16, inset = 5 },
}

-----------------------------------------------------------------------------

local AloftFrame = Aloft:NewModule("Frame", "AceEvent-2.0")
AloftFrame.db = Aloft:AcquireDBNamespace("frame")
AloftFrame.healthBarDB = Aloft:AcquireDBNamespace("healthBar")
Aloft:RegisterDefaults("frame", "profile", {
	enableMouse		= true,
	borderStyle		= "NONE",
	borderColor		= { 1, 1, 1, 1 },
	width			= 10,
	height			= 10,
	packingWidth	= 10,
	packingHeight	= 10,
	backgroundAlpha	= 0,
	colorFormat		= "",
	backgroundColors =
	{
		friendlyPlayer	= { 0, 0, 0 },
		friendlyPet		= { 0, 0, 0 },
		friendlyNPC		= { 0, 0, 0 },
		friendlyBoss	= { 0, 0, 0 },
		hostilePlayer	= { 0, 0, 0 },
		hostilePet		= { 0, 0, 0 },
		hostileNPC		= { 0, 0, 0 },
		hostileBoss		= { 0, 0, 0 },
		neutral			= { 0, 0, 0 },
		pet				= { 0, 0, 0 },
		groupPet		= { 0, 0, 0 },
	},
})

Aloft.Options.args.frame =
{
	type = 'group',
	name = L["Frame"],
	desc = L["Frame options"],
	disabled = function() return not Aloft:IsActive() end,
	args =
	{
		enableMouse  =
		{
			type = 'toggle',
			name = L["Enable Mouse"],
			desc = L["Allows left click to target and right click to interact"],
			get = function() return AloftFrame.db.profile.enableMouse end,
			set = function(v) AloftFrame.db.profile.enableMouse = v AloftFrame:UpdateAll() end,
		},
		width =
		{
			type = 'range',
			name = L["Width"],
			desc = L["Sets the width of the frame"],
			min = 0,
			max = 100,
			step = 1,
			get = function() return AloftFrame.db.profile.width end,
			set = function(v) if AloftFrame.db.profile.width ~= v then AloftFrame.db.profile.width = v AloftFrame.db.profile.packingWidth = v AloftFrame:UpdateAll() end end,
		},
		height =
		{
			type = 'range',
			name = L["Height"],
			desc = L["Sets the height of the frame"],
			min = 0,
			max = 100,
			step = 1,
			get = function() return AloftFrame.db.profile.height end,
			set = function(v) if AloftFrame.db.profile.height ~= v then AloftFrame.db.profile.height = v AloftFrame.db.profile.packingHeight = v AloftFrame:UpdateAll() end end,
		},
		packingWidth =
		{
			type = 'range',
			name = L["Packing Width"],
			desc = L["Sets the minimum horizontal distance between frames"],
			min = 0,
			max = 100,
			step = 1,
			get = function() return AloftFrame.db.profile.packingWidth end,
			set = function(v) AloftFrame.db.profile.packingWidth = v AloftFrame:UpdateAll() end,
		},
		packingHeight =
		{
			type = 'range',
			name = L["Packing Height"],
			desc = L["Sets the minimum vertical distance between frames"],
			min = 0,
			max = 100,
			step = 1,
			get = function() return AloftFrame.db.profile.packingHeight end,
			set = function(v) AloftFrame.db.profile.packingHeight = v AloftFrame:UpdateAll() end,
		},
		backgroundAlpha =
		{
			type = 'range',
			name = L["Background Alpha"],
			desc = L["Sets the alpha for the background of the nameplates"],
			min = 0,
			max = 1,
			step = 0.01,
			get = function() return AloftFrame.db.profile.backgroundAlpha end,
			set = function(v) 
				local oldShow = (AloftFrame.db.profile.backgroundAlpha ~= 0)
				AloftFrame.db.profile.backgroundAlpha = v
				local newShow = (v ~= 0)
				if oldShow ~= newShow then
					Aloft:DetermineDataSources()
				end
				AloftFrame:UpdateAll() 
			end,
		},
		backgroundColors =
		{
			type = 'group',
			name = L["Background Colors"],
			desc = L["Customize colors"],
			disabled = function() return AloftFrame.db.profile.backgroundAlpha == 0 end,
			args =
			{
				colorByClass =
				{
					type = 'toggle',
					name = L["Color by Class"],
					desc = L["Sets the background color according to class"],
					get = function() return AloftFrame.db.profile.colorByClass end,
					set = function(v) AloftFrame.db.profile.colorByClass = v Aloft:DetermineDataSources() AloftFrame:UpdateAll() end,
					order = 1
				},
				colorHostileByClass =
				{
					type = 'toggle',
					name = L["Color Hostile by Class"],
					desc = L["Sets the background color of hostile players according to class"],
					get = function() return AloftFrame.db.profile.colorHostileByClass end,
					set = function(v) AloftFrame.db.profile.colorHostileByClass = v Aloft:DetermineDataSources() AloftFrame:UpdateAll() end,
					order = 2
				},
				setAll =
				{
					type = 'color',
					name = L["Set All"],
					desc = L["Sets the background color for all types"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.friendlyPlayer) end,
					set = function(r, g, b) 
						AloftFrame.db.profile.backgroundColors.friendlyPlayer	= { r, g, b }
						AloftFrame.db.profile.backgroundColors.friendlyPet		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.friendlyNPC		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.friendlyBoss		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.hostilePlayer	= { r, g, b }
						AloftFrame.db.profile.backgroundColors.hostilePet		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.hostileNPC		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.hostileBoss		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.neutral			= { r, g, b }
						AloftFrame.db.profile.backgroundColors.pet				= { r, g, b }
						AloftFrame.db.profile.backgroundColors.groupPet			= { r, g, b }
						Aloft:DetermineDataSources() 
						AloftFrame:UpdateAll() end,
					order = 3
				},
				setAllFriendly =
				{
					type = 'color',
					name = L["Set All Friendly"],
					desc = L["Sets the background color for all friendly types"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.friendlyPlayer) end,
					set = function(r, g, b) 
						AloftFrame.db.profile.backgroundColors.friendlyPlayer	= { r, g, b }
						AloftFrame.db.profile.backgroundColors.friendlyPet		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.friendlyNPC		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.friendlyBoss		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.pet				= { r, g, b }
						AloftFrame.db.profile.backgroundColors.groupPet			= { r, g, b }
						Aloft:DetermineDataSources() 
						AloftFrame:UpdateAll() end,
					order = 4
				},
				setAllHostile =
				{
					type = 'color',
					name = L["Set All Hostile"],
					desc = L["Sets the background color for all hostile types"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.hostileNPC) end,
					set = function(r, g, b) 
						AloftFrame.db.profile.backgroundColors.hostilePlayer	= { r, g, b }
						AloftFrame.db.profile.backgroundColors.hostilePet		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.hostileNPC		= { r, g, b }
						AloftFrame.db.profile.backgroundColors.hostileBoss		= { r, g, b }
						Aloft:DetermineDataSources() 
						AloftFrame:UpdateAll() end,
					order = 5
				},
				friendlyPlayer =
				{
					type = 'color',
					name = L["Friendly Players"],
					desc = L["Sets the background color for friendly players"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.friendlyPlayer) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.friendlyPlayer = { r, g, b } AloftFrame:UpdateAll() end
				},
				friendlyPet =
				{
					type = 'color',
					name = L["Friendly Pets"],
					desc = L["Sets the background color for friendly pets"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.friendlyPet) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.friendlyPet = { r, g, b } Aloft:DetermineDataSources() AloftFrame:UpdateAll() end
				},
				friendlyNPC =
				{
					type = 'color',
					name = L["Friendly NPCs"],
					desc = L["Sets the background color for friendly NPCs"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.friendlyNPC) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.friendlyNPC = { r, g, b } Aloft:DetermineDataSources() AloftFrame:UpdateAll() end
				},
				friendlyBoss =
				{
					type = 'color',
					name = L["Friendly Boss"],
					desc = L["Sets the background color for friendly NPC Bosses"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.friendlyBoss) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.friendlyBoss = { r, g, b } AloftFrame:UpdateAll() end
				},
				hostilePlayer =
				{
					type = 'color',
					name = L["Hostile Players"],
					desc = L["Sets the background color for hostile players"],
					disabled = function() return AloftFrame.db.profile.colorHostileByClass end,
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.hostilePlayer) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.hostilePlayer = { r, g, b } Aloft:DetermineDataSources() AloftFrame:UpdateAll() end
				},
				hostilePet  =
				{
					type = 'color',
					name = L["Hostile Pets"],
					desc = L["Sets the background color for hostile pets"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.hostilePet) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.hostilePet = { r, g, b } Aloft:DetermineDataSources() AloftFrame:UpdateAll() end
				},
				hostileNPC =
				{
					type = 'color',
					name = L["Hostile NPCs"],
					desc = L["Sets the background color for hostile units"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.hostileNPC) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.hostileNPC = { r, g, b } Aloft:DetermineDataSources() AloftFrame:UpdateAll() end
				},
				hostileBoss =
				{
					type = 'color',
					name = L["Hostile Boss"],
					desc = L["Sets the background color for hostile boss units"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.hostileBoss) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.hostileBoss = { r, g, b } AloftFrame:UpdateAll() end
				},
				neutral =
				{
					type = 'color',
					name = L["Neutral Units"],
					desc = L["Sets the background color for neutral units"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.neutral) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.neutral = { r, g, b } AloftFrame:UpdateAll() end
				},
				pet =
				{
					type = 'color',
					name = L["Pet"],
					desc = L["Sets the background color for your pet"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.pet) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.pet = { r, g, b } AloftFrame:UpdateAll() end
				},
				groupPet =
				{
					type = 'color',
					name = L["Group Pets"],
					desc = L["Sets the background color for group pets"],
					get = function() return unpack(AloftFrame.db.profile.backgroundColors.groupPet) end,
					set = function(r, g, b) AloftFrame.db.profile.backgroundColors.groupPet = { r, g, b } Aloft:DetermineDataSources() AloftFrame:UpdateAll() end
				},
				advanced  =
				{
					type = 'group',
					name = L["Advanced"],
					desc = L["Options for expert users"],
					order = -1,
					args =
					{
						format = 
						{
							type = 'text',
							name = L["Color Format"],
							desc = L["Color tag"],
							usage = L["<Any tag string>"],
							get = function() return AloftFrame.db.profile.colorFormat end,
							set = function(v) AloftFrame.db.profile.colorFormat = v Aloft:DetermineDataSources() AloftFrame:UpdateAll() end,
						}
					},
				},
			}
		},
		borderColor =
		{
			type = 'color',
			name = L["Border Color"],
			desc = L["Sets the border color of the nameplate"],
			get = function() return unpack(AloftFrame.db.profile.borderColor) end,
			set = function(r, g, b, a) AloftFrame.db.profile.borderColor = { r, g, b, a } AloftFrame:UpdateAll() end,
			disabled = function() return AloftFrame.db.profile.borderStyle == "NONE" end,
			hasAlpha = true,
		},
		borderStyle =
		{
			type = 'text',
			name = L["Border Style"],
			desc = L["Sets the style of the nameplate border"],
			get = function() return AloftFrame.db.profile.borderStyle end,
			set = function(v) AloftFrame.db.profile.borderStyle = v AloftFrame:UpdateAll() end,
			validate = { ["NONE"] = L["NONE"], ["STANDARD"] = L["STANDARD"], ["SQUARE"] = L["SQUARE"] }
		},
	},
}

-----------------------------------------------------------------------------

local colorMethod = nil
local colorMethodData = { }
local frameUpdateList = { }

-----------------------------------------------------------------------------

function AloftFrame:UpdateAll()
	for aloftData in Aloft:IterateNameplates() do
		self:SetupFrame(aloftData)
	end
	
	for aloftData in Aloft:IterateVisibleNameplates() do
		if aloftData.backdropFrame then
			self:OnNameplateShowWithColor(aloftData)
		else
			self:OnNameplateShow(aloftData)
		end
	end
end

function AloftFrame:RegisterEvents()
	self:UnregisterAllEvents()
	self:RegisterEvent("Aloft:SetupFrame", "SetupFrame")
	self:RegisterEvent("Aloft:OnNameplateHide", "OnNameplateHide")
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "OnCombatFinished")

	local profile = self.db.profile
	local onShowMethod = "OnNameplateShow"
	
	if (profile.backgroundAlpha ~= 0) and
	   (not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.friendlyPet) or
		not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.friendlyNPC) or
		not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.friendlyBoss) or
		not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.hostilePlayer) or
		not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.hostilePet) or
		not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.hostileNPC) or
		not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.hostileBoss) or
		not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.neutral) or
		not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.pet) or
		not self:IsArrayEqual(profile.backgroundColors.friendlyPlayer, profile.backgroundColors.groupPet) or
		profile.colorByClass or
		profile.colorHostileByClass or
		profile.colorFormat ~= "") then
		
		if not profile.colorByClass and
		   not profile.colorHostileByClass and
		   (self:IsArrayEqual(profile.backgroundColors.friendlyNPC, profile.backgroundColors.friendlyPet) and
			self:IsArrayEqual(profile.backgroundColors.friendlyNPC, profile.backgroundColors.friendlyBoss) and
			self:IsArrayEqual(profile.backgroundColors.friendlyNPC, profile.backgroundColors.pet) and
			self:IsArrayEqual(profile.backgroundColors.friendlyNPC, profile.backgroundColors.groupPet)) and
		   (self:IsArrayEqual(profile.backgroundColors.hostilePlayer, profile.backgroundColors.hostilePet) and
			self:IsArrayEqual(profile.backgroundColors.hostilePlayer, profile.backgroundColors.hostileNPC) and
			self:IsArrayEqual(profile.backgroundColors.hostilePlayer, profile.backgroundColors.hostileBoss)) then
			onShowMethod = "OnNameplateShowWithColorByType"
		else		
			onShowMethod = "OnNameplateShowWithColor"
		end
	end

	self:RegisterEvent("Aloft:OnNameplateShow", onShowMethod)

	if profile.backgroundAlpha ~= 0 then	
		if profile.colorByClass or profile.colorHostileByClass then
			self:RegisterEvent("Aloft:OnClassDataChanged", onShowMethod)
		end

		if not self:IsArrayEqual(profile.backgroundColors.friendlyPet, profile.backgroundColors.friendlyNPC) or
		   not self:IsArrayEqual(profile.backgroundColors.hostilePet, profile.backgroundColors.hostileNPC) then
			self:RegisterEvent("Aloft:OnIsPetDataChanged", onShowMethod)
		end
		for events in pairs(colorMethodData.events) do
			self:RegisterEvent(events, onShowMethod)
		end
	end
	
end

-----------------------------------------------------------------------------

function AloftFrame:OnEnable()
	self:UpdateAll()
end

function AloftFrame:OnDisable()
	self:UnregisterAllEvents()

	for aloftData in Aloft:IterateNameplates() do
		local backdropFrame = aloftData.backdropFrame
		if backdropFrame then
			backdropFrame:Hide()
		end
	end
end

local dataRequiredList = { }
function AloftFrame:RequiresData()
	local profile = self.db.profile

	colorMethodData = AloftTag:CreateTag(profile.colorFormat)
	colorMethod = colorMethodData.method
	
	self:RegisterEvents()

	for i = 1,#dataRequiredList do
		dataRequiredList[i] = nil
	end

	if profile.backgroundAlpha ~= 0 then
		if profile.colorByClass or
		   profile.colorHostileByClass or
		   not self:IsArrayEqual(profile.backgroundColors.hostileNPC, profile.backgroundColors.hostilePlayer) then
			table.insert(dataRequiredList, "class")
		end

		if not self:IsArrayEqual(profile.backgroundColors.friendlyPet, profile.backgroundColors.friendlyNPC) or
		   not self:IsArrayEqual(profile.backgroundColors.hostilePet, profile.backgroundColors.hostileNPC) then
			table.insert(dataRequiredList, "isPet")
		end

		if Aloft:IsDataAvailable("unitid") and not self:IsArrayEqual(profile.backgroundColors.friendlyPet, profile.backgroundColors.groupPet) then
			table.insert(dataRequiredList, "unitid")
		end

		for data in pairs(colorMethodData.data) do
			table.insert(dataRequiredList, data)
		end
	end

	return unpack(dataRequiredList)
end

local backdropTable =
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
	insets = { }
}

function AloftFrame:SetupFrame(aloftData)
	--
	-- Using the inbuilt frame for borders/background works fine, until a nameplate shows during combat
	--
	-- The game likes to automatically resize the frame when a frame is shown, 
	-- but we can't change it back OnShow because of in-combat lockdown.
	--
	-- This forces us to use our own backdropFrame which we can control as we please.
	--
	-- We also can't use SetFrameLevel during combat.
	--
	-- Requirements:
	--  Backdrop must appear behind all elements
	--  CastBar/HealthBar must be show backgrounds
	--  CastBar/HealthBar background must be above backdrop
	--  Overlay must display on top of HealthBar/CastBar
	--  UI elements must appear on top of healthbar/castbar/overlay
	--
	local profile = self.db.profile
	local backdropFrame = aloftData.backdropFrame

	if InCombatLockdown() then
		frameUpdateList[aloftData] = true
	else
		aloftData.nameplateFrame:EnableMouse(not self.db.profile.enableMouse)
	end

	if profile.borderStyle == "NONE" and profile.backgroundAlpha == 0 then
		if backdropFrame then
			backdropFrame:Hide()
		end
		return
	end
	
	if not backdropFrame then
		local nameplateFrame = aloftData.nameplateFrame
		backdropFrame = CreateFrame("Frame", nil, nameplateFrame)
		backdropFrame:SetPoint("CENTER", nameplateFrame, "CENTER")
		backdropFrame:SetAlpha(nameplateFrame:GetAlpha())
		backdropFrame:SetFrameLevel(0)
		aloftData.backdropFrame = backdropFrame
	end

	local data = borderData[profile.borderStyle]
	backdropTable.edgeFile		= data.texture
	backdropTable.edgeSize		= data.size
	backdropTable.insets.left	= data.inset
	backdropTable.insets.right	= data.inset
	backdropTable.insets.top	= data.inset
	backdropTable.insets.bottom = data.inset
	
	backdropFrame:SetBackdrop(backdropTable)
	backdropFrame:SetBackdropBorderColor(unpack(profile.borderColor))
	
	local healthBarProfile = self.healthBarDB.profile
	backdropFrame:SetWidth(profile.width+healthBarProfile.width)
	backdropFrame:SetHeight(profile.height+healthBarProfile.height)
	backdropFrame:Show()
	
	-- This shifts the backdrop to border layer, enabling it to display, despite being at frame level 0
	backdropFrame:GetRegions():SetDrawLayer("BORDER")

	self:SetBackgroundColor(aloftData, profile.backgroundColors.friendlyPlayer)
end

function AloftFrame:SetBackgroundColor(aloftData, color)
	aloftData.backdropFrame:SetBackdropColor(color[1], color[2], color[3], self.db.profile.backgroundAlpha)
end

function AloftFrame:SetBackgroundColorRgb(aloftData, color)
	aloftData.backdropFrame:SetBackdropColor(color.r, color.g, color.b, self.db.profile.backgroundAlpha)
end

function AloftFrame:OnNameplateShow(aloftData)
	if InCombatLockdown() then
		frameUpdateList[aloftData] = true
	else
		local profile = self.db.profile
		local healthBarProfile = self.healthBarDB.profile
		local nameplateFrame = aloftData.nameplateFrame

		nameplateFrame:SetWidth(profile.packingWidth+healthBarProfile.width)
		nameplateFrame:SetHeight(profile.packingHeight+healthBarProfile.height)
	end
end

function AloftFrame:OnNameplateHide(aloftData)
	frameUpdateList[aloftData] = nil
end

function AloftFrame:OnCombatFinished()
	for aloftData in pairs(frameUpdateList) do
		aloftData.nameplateFrame:EnableMouse(not self.db.profile.enableMouse)
		self:OnNameplateShow(aloftData)
		frameUpdateList[aloftData] = nil
	end
end

-----------------------------------------------------------------------------

function AloftFrame:ColourFriendlyPlayer(aloftData)
	if self.db.profile.colorByClass then
		local class = aloftData.class
		if class then
			local color = RAID_CLASS_COLORS[class]
			if color then
				self:SetBackgroundColorRgb(aloftData, color)
				return
			end
		end
	end
	self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.friendlyPlayer)
end

function AloftFrame:ColourFriendlyNPC(aloftData)
	-- Could be player pet, group pet, friendly pet, friendlyBoss, or friendlyNPC
	if aloftData.name == UnitName("pet") then
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.pet)
	elseif aloftData.unitid then
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.groupPet)
	elseif aloftData.isPet then
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.friendlyPet)
	elseif aloftData.isBoss then
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.friendlyBoss)
	else
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.friendlyNPC)
	end
end

function AloftFrame:ColourHostile(aloftData)
	-- player, hostile pet, hostile boss, hostileNPC
	if self.db.profile.colorHostileByClass then
		local class = aloftData.class
		if class then
			local color = RAID_CLASS_COLORS[class]
			if color then
				self:SetBackgroundColorRgb(aloftData, color)
				return
			end
		end
	end

	if aloftData.class then
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.hostilePlayer)
	elseif aloftData.isPet then
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.hostilePet)
	elseif aloftData.isBoss then
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.hostileBoss)
	else
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.hostileNPC)
	end
end

function AloftFrame:ColourNeutral(aloftData)
	self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors.neutral)
end

function AloftFrame:ColourUnknown(aloftData)
end

local backgroundColorMethods =
{
	friendlyPlayer = AloftFrame.ColourFriendlyPlayer,
	friendlyNPC = AloftFrame.ColourFriendlyNPC,
	hostile = AloftFrame.ColourHostile,
	neutral = AloftFrame.ColourNeutral,
	unknown = AloftFrame.ColourUnknown,
}

function AloftFrame:OnNameplateShowWithColor(aloftData)
	self:OnNameplateShow(aloftData)
	local override = colorMethod(aloftData)
	if #override >= 6 then
		local rhex, ghex, bhex = override:sub(1, 2), override:sub(3, 4), override:sub(5, 6)
		aloftData.backdropFrame:SetBackdropColor(tonumber(rhex, 16)/255, tonumber(ghex, 16)/255, tonumber(bhex, 16)/255, self.db.profile.backgroundAlpha)
	else
		backgroundColorMethods[aloftData.type](self, aloftData)
	end
end

local backgroundTypeColorLookup =
{
	friendlyPlayer = "friendlyPlayer",
	friendlyNPC = "friendlyNPC",
	hostile = "hostileNPC",
	neutral = "neutral",
	unknown = "friendlyNPC",
}

function AloftFrame:OnNameplateShowWithColorByType(aloftData)
	self:OnNameplateShow(aloftData)

	local override = colorMethod(aloftData)
	if #override >= 6 then
		local rhex, ghex, bhex = override:sub(1, 2), override:sub(3, 4), override:sub(5, 6)
		aloftData.backdropFrame:SetBackdropColor(tonumber(rhex, 16)/255, tonumber(ghex, 16)/255, tonumber(bhex, 16)/255, self.db.profile.backgroundAlpha)
	else
		self:SetBackgroundColor(aloftData, self.db.profile.backgroundColors[backgroundTypeColorLookup[aloftData.type]])
	end
end

-----------------------------------------------------------------------------
